#ifndef WEAK_CLASSIFIER_GENERATOR_H
#define WEAK_CLASSIFIER_GENERATOR_H

#include "weak_classifier.h"
#include <vector>

namespace mimas{

class feature_prototype
{
 public:
  //check the boundary and 
  //calculate references on the image to pick
  //given a base rectangle and size of the
  //image. Return a vector with no element if the bondaries
  //are overcomed
  //the logic of the prototype is encoded here.

  
  //!!!!!use a const reference to avoid useless copy
  virtual std::vector<corner_reference> calculate_feature(int x, int y, int w, int h,
							  int im_w, int im_h) = 0;
};

//!!Research, automate the generation of feature_prototype according to an
//xml file. This may lead to the discovery of interesting features (relative to an
//application domain).

//Description:
//the rectangle counted positively and the same rectangle on its right counted
//negatively

class vertical_edge:public feature_prototype
{
  std::vector<corner_reference> calculate_feature(int x, int y, int w, int h, int im_w, int im_h)
    {
      //      std::vector<corner_reference> & allocated = *(new std::vector<corner_reference>);
      std::vector<corner_reference> allocated;
      
      //check bounds
      if (!(x >=0 && y>=0 && x + 2 * w < im_w && y + h < im_h))
	return allocated;

      allocated.push_back(corner_reference(x, y, 1));
      allocated.push_back(corner_reference(x + w, y, -2));
      allocated.push_back(corner_reference(x + 2 * w, y, 1));
      allocated.push_back(corner_reference(x, y + h, -1));
      allocated.push_back(corner_reference(x + w, y + h, 2));
      allocated.push_back(corner_reference(x + 2 * w, y + h, -1));
      
      return allocated;	
    }
};

//Description:
//the rectangle counted positively and the same rectangle on its bottom counted
//negatively

class horizontal_edge:public feature_prototype
{
  std::vector<corner_reference> calculate_feature(int x, int y, int w, int h, int im_w, int im_h)
    {
      //      std::vector<corner_reference> & allocated = *(new std::vector<corner_reference>);
      std::vector<corner_reference> allocated;
      
      //check bounds
      if (!(x >=0 && y>=0 && x + w < im_w && y + 2 * h < im_h))
	return allocated;

      allocated.push_back(corner_reference(x, y, 1));
      allocated.push_back(corner_reference(x + w, y, -1));
      allocated.push_back(corner_reference(x , y + h, -2));
      allocated.push_back(corner_reference(x + w, y + h, 2));
      allocated.push_back(corner_reference(x , y + 2 * h, 1));
      allocated.push_back(corner_reference(x + w, y + 2 * h, -1));
      return allocated;	
    }
};

//Description:
//the rectangle counted positively and twice the same rectangle on its right counted
//negatively for the first and positively for the second.

class vertical_line:public feature_prototype
{
  std::vector<corner_reference> calculate_feature(int x, int y, int w, int h, int im_w, int im_h)
    {
      //std::vector<corner_reference> & allocated = *(new std::vector<corner_reference>);
      std::vector<corner_reference> allocated;
      //check bounds
      if (!(x >=0 && y>=0 && x + 3 * w < im_w && y + h < im_h))
	return allocated;

      allocated.push_back(corner_reference(x, y, .5));
      allocated.push_back(corner_reference(x + w, y, -1.5));
      allocated.push_back(corner_reference(x + 2 * w, y , 1.5));
      allocated.push_back(corner_reference(x + 3 * w, y , -.5));
      allocated.push_back(corner_reference(x , y + h, -.5));
      allocated.push_back(corner_reference(x + w, y + h, 1.5));
      allocated.push_back(corner_reference(x + 2 * w, y + h, -1.5));
      allocated.push_back(corner_reference(x + 3 * w, y + h, .5));
      return allocated;	
    }
};

 //Description:
//the rectangle counted positively and twice the same rectangle on its bottom counted
//negatively for the first and positively for the second.

class horizontal_line:public feature_prototype
{
  std::vector<corner_reference> calculate_feature(int x, int y, int w, int h, int im_w, int im_h)
    {
      std::vector<corner_reference> allocated;
      //std::vector<corner_reference> & allocated = *(new std::vector<corner_reference>);
      //check bounds
      if (!(x >=0 && y>=0 && x + w < im_w && y + 3 * h < im_h))
	return allocated;

      allocated.push_back(corner_reference(x, y, .5));
      allocated.push_back(corner_reference(x + w, y, -.5));
      allocated.push_back(corner_reference(x , y + h , -1.5));
      allocated.push_back(corner_reference(x + w, y + h , 1.5));
      allocated.push_back(corner_reference(x , y + 2 * h, 1.5));
      allocated.push_back(corner_reference(x + w, y + 2 * h, -1.5));
      allocated.push_back(corner_reference(x , y + 3 * h, -.5));
      allocated.push_back(corner_reference(x + w, y + 3 * h, .5));
      return allocated;	
    }
};

class weak_classifier_generator
{
 public:
  std::vector<weak_classifier> generate(int image_w, int image_h)
    {
      std::vector<weak_classifier> weak_classifiers;
      //calculate all possible rectangle
      //we don't consider rectangle of area 1 pixel
      for(it = feature_prototypes.begin(); it != feature_prototypes.end(); it++)
	for(int w = 2; w <= image_w; w++)
	  for(int h = 2; h <= image_h; h++) 
	    for(int x = 0;(x + w) < image_w; x++)
	      for(int y = 0; (y + h)< image_h ; y++)
		{
		  std::vector<corner_reference> allocated = (*it)->calculate_feature(x, y, w, h, image_w, image_h);
		  if(allocated.size() !=0)
		    {
		      haar_feature allocated2(allocated);
		      weak_classifier temp3(allocated2);
		      weak_classifiers.push_back(temp3);
		    }
		  //	  else
		  //delete &allocated;
		}
      std::cerr <<"Number of classifier generated: "<< weak_classifiers.size() <<std::endl;
      return weak_classifiers;
    }

  void add_feature_prototype(feature_prototype &fp)
    {
      feature_prototypes.push_back(&fp);
    }
  
 private:
  std::vector<feature_prototype *> feature_prototypes;
  std::vector<feature_prototype *>::iterator it;
  
};

}
#endif
