#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <vector>
#include "corrcoef.h"
#include "linalg.h"
#include "mimasexception.h"

using namespace boost;
using namespace boost::numeric;
using namespace std;
using namespace mimas;

typedef double Real;

typedef ublas::matrix< Real, ublas::column_major > Matrix;
typedef ublas::triangular_matrix 
  < Real, ublas::upper, ublas::column_major > TriangularMatrix;
typedef ublas::symmetric_matrix 
  < Real, ublas::upper, ublas::column_major > SymmetricMatrix;
typedef ublas::banded_matrix 
  < Real, ublas::column_major > BandedMatrix;

int main( int, char *[] )
{
  int retVal = 0;
  try {

    cout << fixed << setprecision( 2 );

    { // Correlation-coefficient.
      const int
        dataA[ 5 ] = {  1,  4,  3,  4,  5 },
        dataB[ 5 ] = {  0, -3, -2, -1, -6 };
      
        vector<int>
          a( dataA, dataA + 5 ),
          b( dataB, dataB + 5 );
      
        corrcoef< vector< int > > corrcoef;
        
        cout << "corrcoef( a, b ) = " << corrcoef( a, b ) << endl << endl;
    };

    cout << fixed << setprecision( 2 );

    { // Matrix inversion.
      Matrix m( 2, 2 );
      m( 0, 0 ) =  cos( M_PI / 6 );
      m( 0, 1 ) = -sin( M_PI / 6 );
      m( 1, 0 ) =  sin( M_PI / 6 );
      m( 1, 1 ) =  cos( M_PI / 6 );
      m *= 2.0;
      Matrix mi( inv( m ) );
      cout << "Matrix inversion : M * M^-1 = I" << endl
	   << m << " * " << mi << " = " << prod( m, mi ) << endl << endl;
    };

    cout << fixed << setprecision( 0 );

    { // Cholesky decomposition : A = U^t * U
      // ex :
      //     [ 4 2 0 ]   [ 2 0 0 ]   [ 2 1 0 ]
      // A = [ 2 5 2 ] = [ 1 2 0 ] * [ 0 2 1 ]
      //     [ 0 2 5 ]   [ 0 1 2 ]   [ 0 0 2 ]
     
      cout << "Cholesky decomposition : A = U^t * U" << endl;

      SymmetricMatrix A(3,3);
      TriangularMatrix U;

      A(0,0) = 4;
      A(0,1) = 2;
      A(0,2) = 0;
      A(1,1) = 5;
      A(1,2) = 2;
      A(2,2) = 5;

      cout << "A: " << endl;
      for (unsigned i = 0; i < A.size1 (); ++ i) {
	for (unsigned j = 0; j < A.size2 (); ++ j)
	  cout << A(i, j) << "  ";
	cout << endl;
      }
      cout << endl;
  
      // Call LAPACK algorithm
      U = pptrf( A );
  
      cout << "U: " << endl;
      for (unsigned i = 0; i < U.size1 (); ++ i) {
	for (unsigned j = 0; j <  U.size2 (); ++ j)
	  cout << ((i<j) ? U(i, j) : U(j, i)) << "  ";
	cout << endl;
      }
      cout << endl;
    };

    cout << fixed << setprecision( 2 );

    { // Singular value decomposition:
      Matrix A( 3, 2 ), U, Vt;
      
      A( 0, 0 ) =  1; A( 0, 1 ) =  2;
      A( 1, 0 ) = -1; A( 1, 1 ) = -1;
      A( 2, 0 ) =  2; A( 2, 1 ) =  3;

      // According to the documentation 'gesvd' is faster than 'gesdd'
      // for small matrices.
      BandedMatrix S( gesvd( A, &U, &Vt ) );

      // Nested function calls to 'prod' are not working on some systems.
      Matrix tmp( prod( U, S ) );

      cout << "Singular value decomposition: A = U * S * V^t = A" << endl
	   << A << " = " << U << " * " << S << " * " << Vt << " = "
	   << prod( tmp, Vt ) << endl << endl;
    };


  } catch ( exception &e ) {
    cerr << e.what() << endl;
    retVal = 1;
  };
  return retVal;
}
