
#ifndef GNUPLOT_H
#define GNUPLOT_H

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <cstdarg>
#include <vector>
#include "mimasexception.h"
#include "image.h"

namespace mimas {
/**
 * Gnuplot access functions
 * See <A HREF="info:/gnuplot">gnuplot info-pages</A>.
 * @author Bala Amavasai (bala@amavasai.org)
 * @date Mon May 13 12:45:41 BST 2002
 * $Header: /cvs/mimas2/include/gnuplot.h,v 1.1.1.1 2005/08/09 15:37:45 engmb Exp $
 */
class gnuplot : public object
{
 public:
  typedef std::string string;
	private:
		void checkOpen(void); ///< open a gnuplot session if one has not been opened already
      std::vector< string > lineTitles;
		// bool linetitlesSet;
		FILE *fp; ///< the pipe used to send commands to a gnuplot session
		
	public:
		gnuplot(void); ///< constructor
		~gnuplot(void); ///< destructor
		void command(const char *); ///< send a const char * command to the gnuplot session. You have to terminate with a \n. This allows for commands to be queued
		void command(const string &s); ///< send a string command to the gnuplot session. You have to terminate with a \n. This allows commands to be queued
        // void plot(matrix_tool& mat) throw(mimasexception); 
      /// Plot 1-D matrix - like in Octave
      template< typename T >
      void plot( const boost::numeric::ublas::vector< T > &vec );
      template< typename T >
      void plot( const boost::numeric::ublas::matrix< T > &mat );
      /** Plotting of xcol against ycol.
          Columns are indexed from 0. style is either "line" or "point". type
          is an integer which specifies the type of line or point (refer to
          gnuplot package). size is the point or line size. Use setMultiPlot
          to overlay plots. */
      template< typename T >
      void plot( const boost::numeric::ublas::matrix< T > &mat,
                 int xcol, int ycol,
                 const std::string &style = "points",
                 int type=1, int size=1 ) throw (mimasexception);
      /** plotting of x and single or multple y column pairs.
          List of y columns must be terminated by NULL or it will lead to
          indiscriminate behaviour. Columns are indexed from 0. */
      template< typename T >
      void plotxy( const boost::numeric::ublas::matrix< T > &mat,
                   int xcol, int ycol, ... ) throw (mimasexception);
      template< typename T >
      void plotImageProfile( const image< T > &im ); ///< plot an image<int> profile
		void open(void); ///< open a gnuplot session
		void close(void); ///< close a gnuplot session
		void startSave(const char *fn, const char *format); ///< start saving future plotting commands to file. The first parameter is the filename. The second parameter is the fileformat and is either one of PS, EPS, PSCOLOR, EPSCOLOR, JPG or PNG. You must end the writing to file procedure by calling endSave()
		void endSave(void); ///< complementary command for startSave().
		void setLineTitle(const char *title, ...); ///< set line titles for use with plotxy
		void setMultiPlot(bool multiplot);///< set or unset multiplot
		void unsetLineTitle(void); ///< unset line titles
		gnuplot& operator<<( const string &s ); ///< pipe strings to the gnuplot session

};

// Example of using gnuplot

#ifdef GNUPLOT_H_EXAMPLE_H
MMGLOBALDEFS

int main()
{
	image<int> im;
	gnuplot graph;

	graph << "set isosamples 40,40";
	graph << "set xrange[-3:3]";
	graph << "set yrange[-3:3]";
	graph << "set zrange[-2:2]";
	graph << "set hid; set nokey";


	for (;;)
	{
		char s[255];

		for (double i=1; i<2; i+=0.05)
		{
		    sprintf(s,"splot cos(x*x*%e+y*y*%e)\n",i,i);
		    graph.command(s);
		}

		for (double i=2; i>1; i-=0.05)
		{
		    sprintf(s,"splot cos(x*x*%e+y*y*%e)\n",i,i);
		    graph.command(s);
		}

	}

	sleep(3);

	return 0;
}
#endif
}

#include "gnuplot.tcc"

#endif
