#ifndef IMAGE_FILEINPUT_H
#define IMAGE_FILEINPUT_H

#include <list>
#include <map>
#include <Magick++.h>
#include "image_input.h"
#include "rgba.h"

namespace mimas {

/** @addtogroup imageIO
    @{ */
/** @addtogroup imageInput
    @{ */
/** Class for reading an image from a stream.
    The stream is used for reading image(s) from.

    The <A HREF="http://www.imagemagick.org/Magick++/">Magick++</A> library
    is used for reading the images.

    Note, that it is only possible to read a stream, which contains a single
    image.

    Currently <A HREF="http://www.imagemagick.org/Magick++/">Magick++</A>
    supports reading the following file-formats: 8BIM 8BIMTEXT 8BIMWTEXT APP1
    APP1JPEG ART AVI AVS BIE BMP CAPTION CMYK CMYKA CUR CUT DCM DCX DPS DPX
    EPDF EPI EPS EPSF EPSI EPT EXIF FAX FITS FPX FRACTAL G3 GIF GIF87 GRADIENT
    GRANITE GRAY ICB ICC ICM ICO ICON IPTC IPTCTEXT IPTCWTEXT JBG JBIG JNG
    JPEG JPG LABEL LOGO M2V MAGICK MAP MAT MIFF MNG MONO MPC MPEG MPG MSL MTV
    MVG NETSCAPE NULL OTB P7 PAL PALM PATTERN PBM PCD PCDS PCT PCX PDB PDF PFA
    FB PGM PICON PICT PIX PLASMA PM PNG PNG24 PNG32 PNG8 PNM PPM PS PSD PTIF
    PWP RAS RGB RGBA RLA RLE SCT SFW SGI STEGANO SUN SVG TEXT TGA TIF TIFF
    TILE TIM TTF TXT UYVY VDA VICAR VID VIFF VST WBMP WMF WPG X XBM XC XCF XMP
    XPM XV XWD YUV

    @author Jan Wedekind (jan@wedesoft.de)
    @date Thu Jan 12 14:19:44 UTC 2006
    @see image_fileoutput
    @see image_filesinput */
template< typename T >
class image_fileinput: public image_input< T >
{
 public:
  /** Constructor.
      The constructor stores a pointer to the specified stream-object.
      @param _stream Stream to read image from later. */
  explicit image_fileinput( std::istream &_stream ): stream( &_stream ) {}
  /** Read image from stream.
      An image is constructed from the data provided by the stream. If there
      are no more images, the state of this object is set to \c false (end of
      stream).
      @param image Object to store image in. */
  virtual void read( image< T > &image ) throw (mimasexception);
  /** Get list of supported file-formats.
      Gets a map of file-formats, which have read-support by Magick++. The
      boolean is indicating, wether the file-formats may contain multiple
      images or not. */
  static std::map< std::string, bool > formats(void);
 protected:
  /** Identification of mapping.
      This method returns a parameter passed to \c Magick::Image::write. The
      return value either is "BGRA" or "I". */
  static const char *mapString(void);
  /// Pointer to stream.
  std::istream *stream;
  /// Buffer for unread images.
  boost::shared_ptr< std::list< Magick::Image > > images;
};

template<>
const char *image_fileinput< unsigned char >::mapString(void);

template<>
const char *image_fileinput< rgba< unsigned char > >::mapString(void);
 
/// Stream-operator to read image from stream.
template< typename T, typename U >
inline image_fileinput< T > &operator>>( image_fileinput< T > &input,
                                            image< U > &im )
{
  if ( typeid( T ) == typeid( U ) )
    input.read( (image< T > &)im );
  else {
    image< T > temp;
    input.read( temp );
    im = image< U >( temp );
  };
  return input;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream, image< unsigned char > &im )
{
  image_fileinput< unsigned char > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< int > &im )
{
  image_fileinput< unsigned char > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< float > &im )
{
  image_fileinput< unsigned char > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< double > &im )
{
  image_fileinput< unsigned char > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< rgba< unsigned char > >
                                 &im )
{
  image_fileinput< rgba< unsigned char > > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< rgba< int > >
                                 &im )
{
  image_fileinput< rgba< unsigned char > > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< rgba< double > >
                                 &im )
{
  image_fileinput< rgba< unsigned char > > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

/// For convenience.
inline std::istream &operator>>( std::istream &stream,
                                 image< rgba< float > >
                                 &im )
{
  image_fileinput< rgba< unsigned char > > input( stream );
  try {
    input >> im;
  } catch ( mimasexception & ) {
    stream.setstate( std::ios_base::failbit );
  };
  return stream;
}

///@}

///@}

};

#include "image_fileinput.tcc"

#endif
