#ifndef __MULTI_ARRAY_OP_TCC
#define __MULTI_ARRAY_OP_TCC

// For compatibility with boost <1.31 when apply_if replaced with eval_if
#include <boost/version.hpp>
#if BOOST_VERSION<=103100
#include <boost/mpl/apply_if.hpp>
#ifndef eval_if
#define eval_if apply_if
#endif
#else
#include <boost/mpl/eval_if.hpp>
#endif

#include <boost/multi_array.hpp>
#include "functions.h"

namespace mimas {

// scalar = f( scalar )
template< class _F >
struct element_apply_intern
{
  typedef element_apply_intern< _F > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  element_apply_intern( const _F &_f ): f(_f) {}
  void operator()( typename _F::argument_type *result,
                   const size_type *,
                   const index_type * ) const {
    *result = f( *result );
  };
  _F f;
};

// array = f( array )
template< class _F, int _N >
struct array_apply_intern
{
  typedef array_apply_intern< _F, _N > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  array_apply_intern( const _F &_f ): f(_f) {}
  void operator()( typename _F::argument_type *result,
                   const size_type *sizeR,
                   const index_type *stepR ) const {
    size_type size = *sizeR;
    index_type step = *stepR;
    for ( int i=0; i<(signed)size; i++ ) {
      f( result, sizeR + 1, stepR + 1 );
      result += step;
    };
  };
  typename boost::mpl::eval_if< boost::mpl::bool_< _N == 1 >,
                                 element_apply_intern< _F >,
                                 array_apply_intern< _F, _N - 1 > >::type f;
};

// f( scalar )
template< class _F >
struct element_apply_extern
{
  typedef element_apply_extern< _F > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  element_apply_extern( const _F &_f ): f(_f) {}
  void operator()( typename _F::result_type *result,
                   const typename _F::argument_type *x,
                   const size_type *,
                   const size_type *,
                   const index_type *,
                   const index_type * ) const {
    *result = f( *x );
  };
  _F f;
};

// f( array )
template< class _F, int _N >
struct array_apply_extern
{
  typedef array_apply_extern< _F, _N > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  array_apply_extern( const _F &_f ): f(_f) {}
  void operator()( typename _F::result_type *result,
                   const typename _F::argument_type *x,
                   const size_type *sizeR,
                   const size_type *sizeX,
                   const index_type *stepR,
                   const index_type *stepX ) const {
    size_type size = *sizeR;
    index_type
      sR = *stepR,
      sX = *stepX;
    for ( int i=0; i<(signed)size; i++ ) {
      f( result, x, sizeR + 1, sizeX + 1, stepR + 1, stepX + 1 );
      result += sR;
      x += sX;
    };
  };
  typename boost::mpl::eval_if< boost::mpl::bool_< _N == 1 >,
                                 element_apply_extern< _F >,
                                 array_apply_extern< _F, _N - 1 > >::type f;
};

// scalar = f( scalar )
template< class _F >
struct element_element_apply_intern
{
  typedef element_element_apply_intern< _F > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  element_element_apply_intern( const _F &_f ): f(_f) {}
  void operator()( typename _F::first_argument_type *result,
                   const typename _F::second_argument_type *x,
                   const size_type *,
                   const size_type *,
                   const index_type *,
                   const index_type * ) const {
    *result = f( *result, *x );
  };
  _F f;
};

// array = f( array )
template< class _F, int _N >
struct array_array_apply_intern
{
  typedef array_array_apply_intern< _F, _N > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  array_array_apply_intern( const _F &_f ): f(_f) {}
  void operator()( typename _F::first_argument_type *result,
                   const typename _F::second_argument_type *x,
                   const size_type *sizeR,
                   const size_type *sizeX,
                   const index_type *stepR,
                   const index_type *stepX ) const {
    size_type size = *sizeR;
    index_type
      stR = *stepR,
      stX = *stepX;
    for ( int i=0; i<(signed)size; i++ ) {
      f( result, x, sizeR + 1, sizeX + 1, stepR + 1, stepX + 1 );
      result += stR;
      x += stX;
    };
  };
  typename boost::mpl::eval_if< boost::mpl::bool_< _N == 1 >,
                                 element_element_apply_intern< _F >,
                                 array_array_apply_intern< _F, _N - 1 > >::type f;
};

// f( scalar )
template< class _F >
struct element_element_apply_extern
{
  typedef element_element_apply_extern< _F > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  element_element_apply_extern( const _F &_f ): f(_f) {}
  void operator()( typename _F::result_type *result,
                   const typename _F::first_argument_type *x,
                   const typename _F::second_argument_type *y,
                   const size_type *,
                   const size_type *,
                   const size_type *,
                   const index_type *,
                   const index_type *,
                   const index_type * ) const {
    *result = f( *x, *y );
  };
  _F f;
};

// f( array )
template< class _F, int _N >
struct array_array_apply_extern
{
  typedef array_array_apply_extern< _F, _N > type;
  typedef boost::detail::multi_array::size_type size_type;
  typedef boost::detail::multi_array::index index_type;
  array_array_apply_extern( const _F &_f ): f(_f) {}
  void operator()( typename _F::result_type *result,
                   const typename _F::first_argument_type *x,
                   const typename _F::second_argument_type *y,
                   const size_type *sizeR,
                   const size_type *sizeX,
                   const size_type *sizeY,
                   const index_type *stepR,
                   const index_type *stepX,
                   const index_type *stepY ) const {
    size_type size = *sizeR;
    index_type
      sR = *stepR,
      sX = *stepX,
      sY = *stepY;
    for ( int i=0; i<(signed)size; i++ ) {
      f( result, x, y,
         sizeR + 1, sizeX + 1, sizeY + 1,
         stepR + 1, stepX + 1, stepY + 1 );
      result += sR;
      x += sX;
      y += sY;
    };
  };
  typename boost::mpl::eval_if< boost::mpl::bool_< _N == 1 >,
                                 element_element_apply_extern< _F >,
                                 array_array_apply_extern< _F, _N - 1 > >::type f;
};

// Element-wise operations on arrays.
// f( array )
#define ARRAY_OP_INTERN( ARRAYTYPE, MODIFIER, INAME, OBJECT )        \
  template< typename T, std::size_t N >                                 \
  ARRAYTYPE< T, N > MODIFIER INAME( ARRAYTYPE< T, N > MODIFIER x )      \
  {                                                                     \
    OBJECT< T > object;                                                 \
    mimas::array_apply_intern< OBJECT< T >, N >                      \
      apply( object );                                                  \
    apply( x.origin(),                                                  \
           x.shape(),                                                   \
           x.strides() );                                               \
    return x;                                                           \
  }

#define ARRAY_OP_EXTERN( ARRAYTYPE, MODIFIER, ENAME, OBJECT )         \
  template< typename T, std::size_t N >                                  \
  boost::multi_array< typename OBJECT< T >::result_type, N >             \
    ENAME( const ARRAYTYPE< T, N > MODIFIER x )                          \
  {                                                                      \
    boost::array< std::size_t, N > shape;                                \
    std::copy( x.shape(), x.shape() + N, shape.begin() );                \
    boost::multi_array< typename OBJECT< T >::result_type, N >           \
      retVal( shape );                                                   \
    OBJECT< T > object;                                                  \
    mimas::array_apply_extern< OBJECT< T >, N >                       \
      apply( object );                                                   \
    apply( retVal.origin(), x.origin(),                                  \
           retVal.shape(), x.shape(),                                    \
           retVal.strides(), x.strides() );                              \
    return retVal;                                                       \
  }

#define ARRAY_OP_INTERN_H( INAME, OBJECT )                                \
  ARRAY_OP_INTERN( boost::detail::multi_array::sub_array             ,  , \
                      INAME, OBJECT )                                        \
  ARRAY_OP_INTERN( boost::multi_array                                , &, \
                      INAME, OBJECT )                                        \
  ARRAY_OP_INTERN( boost::detail::multi_array::multi_array_view      , &, \
                      INAME, OBJECT )

#define ARRAY_OP_EXTERN_H( ENAME, OBJECT )                                \
  ARRAY_OP_EXTERN( boost::detail::multi_array::const_sub_array       ,  , \
                      ENAME, OBJECT )                                        \
  ARRAY_OP_EXTERN( boost::detail::multi_array::sub_array             ,  , \
                      ENAME, OBJECT )                                        \
  ARRAY_OP_EXTERN( boost::multi_array                                , &, \
                      ENAME, OBJECT )                                        \
  ARRAY_OP_EXTERN( boost::detail::multi_array::const_multi_array_view, &, \
                      ENAME, OBJECT )                                        \
  ARRAY_OP_EXTERN( boost::detail::multi_array::multi_array_view      , &, \
                      ENAME, OBJECT )

#define ARRAY_OP( INAME, ENAME, OBJECT )     \
  ARRAY_OP_INTERN_H( INAME, OBJECT )         \
  ARRAY_OP_EXTERN_H( ENAME, OBJECT ) 

#define ARRAY_OP_EX( ENAME, OBJECT )          \
  ARRAY_OP_EXTERN_H( ENAME, OBJECT )

// array = f( array, scalar )
#define ARRAY_ARRAY_OP_INTERN( LEFTTYPE, LEFTMODIFIER,               \
                                  RIGHTTYPE, INAME, OBJECT )            \
  template< typename T, std::size_t N >                                 \
  LEFTTYPE< T, N > LEFTMODIFIER INAME( LEFTTYPE< T, N > LEFTMODIFIER x, \
                                       const RIGHTTYPE< T, N > &y )     \
  {                                                                     \
    OBJECT< T > f;                                                      \
    mimas::array_array_apply_intern< OBJECT< T >, N >                \
      apply( f );                                                       \
    apply( x.origin(), y.origin(),                                      \
           x.shape(), y.shape(),                                        \
           x.strides(), y.strides() );                                  \
    return x;                                                           \
  }

#define ARRAY_ARRAY_OP_EXTERN( LEFTTYPE, RIGHTTYPE, ENAME, OBJECT )  \
  template< typename T, std::size_t N >                                 \
  boost::multi_array< T, N > ENAME( const LEFTTYPE< T, N > &x,          \
                                    const RIGHTTYPE< T, N > &y )        \
  {                                                                     \
    boost::array< int, N > shape;                                       \
    std::copy( x.shape(), x.shape() + N, shape.begin() );               \
    boost::multi_array< T, N >                                          \
      retVal( shape );                                                  \
    OBJECT< T > f;                                                      \
    mimas::array_array_apply_extern< OBJECT< T >, N >                \
      apply( f );                                                       \
    apply( retVal.origin(), x.origin(), y.origin(),                     \
           retVal.shape(), x.shape(), y.shape(),                        \
           retVal.strides(), x.strides(), y.strides() );                \
    return retVal;                                                      \
  }

// f( array1, array2 )
#define ARRAY_ARRAY_OP_CONST_H2( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE, \
                                    INAME, ENAME, OBJECT )             \
  ARRAY_ARRAY_OP_EXTERN( LEFTTYPE, RIGHTTYPE, ENAME, OBJECT )

// array1 = f( array1, array2 ); f( array1, array2 )
#define ARRAY_ARRAY_OP_NONCONST_H2( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE,     \
                                       INAME, ENAME, OBJECT )                 \
  ARRAY_ARRAY_OP_CONST_H2( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE,              \
                              INAME, ENAME, OBJECT )                          \
  ARRAY_ARRAY_OP_INTERN( LEFTTYPE, LEFTMODIFIER, RIGHTTYPE, INAME, OBJECT )

// array1 = f( array1, array2 ); f( array1, array2 )
#define ARRAY_ARRAY_OP_H1( LEFTTYPE, PROTECT, LEFTMODIFIER, INAME, ENAME,  \
                              OBJECT )                                        \
  ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::const_sub_array       , INAME, ENAME, OBJECT )\
  ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::sub_array             , INAME, ENAME, OBJECT )\
  ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::multi_array                                , INAME, ENAME, OBJECT )\
  ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::const_multi_array_view, INAME, ENAME, OBJECT )\
  ARRAY_ARRAY_OP_##PROTECT##_H2                                            \
  ( LEFTTYPE, LEFTMODIFIER,                                                   \
    boost::detail::multi_array::multi_array_view      , INAME, ENAME, OBJECT )

// array1 = f( array1, array2 ); f( array1, array2 )
#define ARRAY_ARRAY_OP( INAME, ENAME, OBJECT )                           \
  ARRAY_ARRAY_OP_H1( boost::detail::multi_array::const_sub_array       , \
                        CONST   ,  , INAME, ENAME, OBJECT )                 \
  ARRAY_ARRAY_OP_H1( boost::detail::multi_array::sub_array             , \
                        NONCONST,  , INAME, ENAME, OBJECT )                 \
  ARRAY_ARRAY_OP_H1( boost::multi_array                                , \
                        NONCONST, &, INAME, ENAME, OBJECT )                 \
  ARRAY_ARRAY_OP_H1( boost::detail::multi_array::const_multi_array_view, \
                        CONST   , &, INAME, ENAME, OBJECT )                 \
  ARRAY_ARRAY_OP_H1( boost::detail::multi_array::multi_array_view      , \
                        NONCONST, &, INAME, ENAME, OBJECT )

// array = f( array, scalar )
#define ARRAY_SCALAR_OP_INTERN( ARRAYTYPE, MODIFIER, INAME,         \
                                    OBJECT )                            \
  template< typename T, std::size_t N >                                 \
  ARRAYTYPE< T, N > MODIFIER INAME( ARRAYTYPE< T, N > MODIFIER x,       \
                                    const T &y )                        \
  {                                                                     \
    OBJECT< T > f;                                                      \
    mimas::array_apply_intern< std::binder2nd                        \
      < OBJECT< T > >, N >                                              \
      apply( std::bind2nd( f, y ) );                                    \
    apply( x.origin(), x.shape(), x.strides() );                        \
    return x;                                                           \
  }

#define ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, ENAME,   \
                                          OBJECT )                      \
  template< typename T, std::size_t N >                                 \
  boost::multi_array< T, N > ENAME( const ARRAYTYPE< T, N > MODIFIER x, \
                                    const T &y )                        \
  {                                                                     \
    boost::array< int, N > shape;                                       \
    std::copy( x.shape(), x.shape() + N, shape.begin() );               \
    boost::multi_array< T, N >                                          \
      retVal( shape );                                                  \
    OBJECT< T > f;                                                      \
    mimas::array_apply_extern< std::binder2nd                        \
      < OBJECT< T > >, N >                                              \
      apply( std::bind2nd( f, y ) );                                    \
    apply( retVal.origin(), x.origin(),                                 \
           retVal.shape(), x.shape(),                                   \
           retVal.strides(), x.strides() );                             \
    return retVal;                                                      \
  }

#define ARRAY_SCALAR_OP_EXTERN_LEFT( ARRAYTYPE, MODIFIER, ENAME,    \
                                         OBJECT )                       \
  template< typename T, std::size_t N >                                 \
  boost::multi_array< T, N > ENAME( const T &x,                         \
                                    const ARRAYTYPE< T, N > MODIFIER y ) \
  {                                                                     \
    boost::array< int, N > shape;                                       \
    std::copy( y.shape(), y.shape() + N, shape.begin() );               \
    boost::multi_array< T, N >                                          \
      retVal( shape );                                                  \
    OBJECT< T > f;                                                      \
    mimas::array_apply_extern< std::binder1st                        \
      < OBJECT< T > >, N >                                              \
      apply( std::bind1st( f, x ) );                                    \
    apply( retVal.origin(), y.origin(),                                 \
           retVal.shape(), y.shape(),                                   \
           retVal.strides(), y.strides() );                             \
    return retVal;                                                      \
  }

// array = f( array, scalar ); f( array, scalar ); f^*( scalar, array )
// with f^* := ( x, y ) -> f( y, x )
#define ARRAY_SCALAR_OP_NOSWAP_NONCONST( ARRAYTYPE, MODIFIER,      \
                                              INAME, ENAME, OBJECT )    \
  ARRAY_SCALAR_OP_INTERN      ( ARRAYTYPE, MODIFIER, INAME, OBJECT )  \
  ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// array = f( array, scalar ); f( array, scalar ); f( scalar, array )
#define ARRAY_SCALAR_OP_SWAP_NONCONST( ARRAYTYPE, MODIFIER,      \
                                           INAME, ENAME, OBJECT )    \
  ARRAY_SCALAR_OP_INTERN      ( ARRAYTYPE, MODIFIER, INAME, OBJECT )  \
  ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )  \
  ARRAY_SCALAR_OP_EXTERN_LEFT ( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// f( array, scalar )
#define ARRAY_SCALAR_OP_NOSWAP_CONST( ARRAYTYPE, MODIFIER,        \
                                           INAME, ENAME, OBJECT )      \
  ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// f( array, scalar ); f( scalar, array )
#define ARRAY_SCALAR_OP_SWAP_CONST( ARRAYTYPE, MODIFIER,            \
                                        INAME, ENAME, OBJECT )          \
  ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )  \
  ARRAY_SCALAR_OP_EXTERN_LEFT ( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// f( array, scalar )
#define ARRAY_SCALAR_OP_NOSWAP_CONST( ARRAYTYPE, MODIFIER,        \
                                          INAME, ENAME, OBJECT )      \
  ARRAY_SCALAR_OP_EXTERN_RIGHT( ARRAYTYPE, MODIFIER, ENAME, OBJECT )

// Branch depending on type of operator ( NOSWAP or SWAP )
#define ARRAY_SCALAR_OP( INAME, ENAME, OBJECT, TYPE )                  \
  ARRAY_SCALAR_OP_##TYPE##_NONCONST                                    \
  ( boost::multi_array                                , &, INAME, ENAME,  \
    OBJECT )                                                              \
  ARRAY_SCALAR_OP_##TYPE##_NONCONST                                    \
  ( boost::detail::multi_array::sub_array             ,  , INAME, ENAME,  \
    OBJECT )                                                              \
  ARRAY_SCALAR_OP_##TYPE##_NONCONST                                    \
  ( boost::detail::multi_array::multi_array_view      , &, INAME, ENAME,  \
    OBJECT )                                                              \
  ARRAY_SCALAR_OP_##TYPE##_CONST                                       \
  ( boost::detail::multi_array::const_sub_array       ,  , INAME, ENAME,  \
    OBJECT )                                                              \
  ARRAY_SCALAR_OP_##TYPE##_CONST                                       \
  ( boost::detail::multi_array::const_multi_array_view, &, INAME, ENAME,  \
    OBJECT )

// Element-wise multiplication
ARRAY_ARRAY_OP( operator*=, operator*, std::multiplies )
// Element-wise division
ARRAY_ARRAY_OP( operator/=, operator/, std::divides    )
// Element-wise addition
ARRAY_ARRAY_OP( operator+=, operator+, std::plus       )
// Element-wise subtraction
ARRAY_ARRAY_OP( operator-=, operator-, std::minus      )
// Sum of squares
ARRAY_ARRAY_OP( sumSquaresIt, sumSquares, _sumsquares )
// Compute angle.
ARRAY_ARRAY_OP( orientationIt, orientation, _orientation )

// Scalar multiplication
ARRAY_SCALAR_OP( operator*=, operator*, std::multiplies, SWAP   )
// Scalar division
ARRAY_SCALAR_OP( operator/=, operator/, std::divides   , SWAP   )
// Scalar addition
ARRAY_SCALAR_OP( operator+=, operator+, std::plus      , SWAP   )
// Scalar subtraction
ARRAY_SCALAR_OP( operator-=, operator-, std::minus     , SWAP   )
// Threshold values.
ARRAY_SCALAR_OP( thresholdIt, threshold, _threshold, NOSWAP )

// Absolute value.
ARRAY_OP( absoluteIt, absolute, _abs )
// Square root.
ARRAY_OP( squareRootIt, squareRoot, _sqrt )

// Elementwise negation.
ARRAY_OP_EX( operator-, std::negate )

};

#endif
