#include "object.h"
#include "mimasexception.h"
#include "matrix_tool.h"
#include "values.h"

#ifndef NN_HEBBIAN_H
#define NN_HEBBIAN_H

namespace mimas {
	/**
	 * Hebbian networks
	 * @author Bala Amavasai (bala@amavasai.org)
	 * @date Tue Jun 29 18:26:06 BST 2004
	 * $Header: /cvs/mimas2/include/nn_hebbian.h,v 1.1.1.1 2005/08/09 15:37:45 engmb Exp $
	 */
  class nn_hebbian: public object
	{

		private:
			matrix_tool weights; ///< weights
			int trainingType; ///< type of training - see setTrainingType() below
			double eta; ///< learning rate

		public:
			nn_hebbian(void); ///< constructor
			~nn_hebbian(void); ///< destructor
			void init(int rows, int cols); ///< create a Hebbian network with defined number rows and columns
			void setTrainingType(int type); ///< set training type to 0=LTP, 1=post-not-pre-LTD, 2=pre-not-post-LTD
			void setLearningRate(double lr); ///< set the learning rate
         void update(matrix_tool& pat, double targetval) throw (exception); ///< update the Hebbian weights given pattern pat
         double recall(matrix_tool& pat) throw (exception); ///< given pattern pat return the predicted output
			void useWeights(matrix_tool& weight_matrix); ///< use the weights given by weight_matrix
			matrix_tool getWeights(void); ///< return the weights that are stored
	};


	nn_hebbian::nn_hebbian()
	{
		trainingType=0; // default training type
		eta=0.01;
	}

	void nn_hebbian::init(int rows, int cols)
	{
		if (weights.initialised()) weights.clear();
		weights.init(rows, cols);
	}

	void nn_hebbian::setTrainingType(int type)
	{
		trainingType=type;
	}

	void nn_hebbian::setLearningRate(double lr)
	{
		eta=lr;
	}

	// update the Hebbian weights given pattern pat
	void nn_hebbian::update(matrix_tool& pat, double targetval)
     throw (exception)
	{
     MMERROR( weights.getRows() ==pat.getRows() &&
              weights.getCols() == pat.getCols(),
              exception, , "nn_hebbian::update - weight and input "
              "matrices dimensions not equal" );

		for (int i=0; i<weights.getRows(); i++)
			for (int j=0; j<weights.getCols(); j++)
			{
				switch (trainingType)
				{
					// *post-not-pre* long term depression (LTD)
					case 1: weights(i,j) = weights(i,j) + eta * (2.0*pat(i,j)-1) * targetval;
							break;

							// *pre-not-post* long term depression (LTD)
					case 2: weights(i,j) = weights(i,j) + eta * pat(i,j) * (2 * targetval -1);
							break;

							// Long term potention (LTP)
					default: weights(i,j) = weights(i,j) + eta * pat(i,j) *targetval;
				}

			}

	}


	double nn_hebbian::recall(matrix_tool& pat) throw (exception)
	{
     MMERROR( weights.getRows() == pat.getRows() &&
              weights.getCols() == pat.getCols(), exception, ,
              "nn_hebbian::recall - weight and input matrices dimensions "
              "not equal" );

		double targetval=0.0;

		for (int i=0; i<weights.getRows(); i++)
			for (int j=0; j<weights.getCols(); j++)
				targetval=targetval + weights(i,j) * pat(i,j);

		return targetval;

	}


	// use the weights given by weight_matrix;
	void nn_hebbian::useWeights(matrix_tool& weight_matrix)
	{
		if (weights.initialised()) weights.clear();
		weights=weight_matrix;
	}


	// return the weights that are stored	
	matrix_tool nn_hebbian::getWeights(void)
	{
		matrix_tool temp=weights;

		return weights;
	}
}

#endif
