#ifndef __WARP_H
#define __WARP_H

#include <boost/numeric/ublas/vector.hpp>
#include "image.h"

namespace mimas {

/** @defgroup warping Image warping
    Mimas-API for warping images.
    @{ */
/** Image to image transform.
    The image \f$g\in\mathrm{R}^2\mapsto\mathrm{T}\f$ is warped using a
    2D-tensor \f$t\in\mathrm{R}^2\mapsto\mathrm{R}^2\f$:
    \f$g^\prime(\vec{x})=g(\vec{x}+t(\vec{x}-\vec{m}))\f$, where \f$\vec{m}\f$
    is the middle of the image.

    You need to define your own tensor function as function object:
    \include lenscorrection/main.cc

    @todo Maybe the vectors passed to the tensor should be scaled, so that the
    tensor-specification will be independent of the image resolution.
    @todo Provide functionality for using interpolation.
    @author Jan Wedekind (jan@wedesoft.de)
    @date Wed Jul 07 14:13:00 UTC 2004
    @see vectorfield  */
template< class Tensor, typename T >
class warp: public object
{
 public:
  typedef boost::numeric::ublas::vector< double > Vector;
  /** Constructor.
      @param aT tensor to apply to images.
      @param aBackground Background value to use for undefined pixels. */
  warp( const Tensor &aT, T aBackground ): t(aT), background(aBackground) {}
  /** Warp image.
      @param in Input image.
      @return Output image. */
  mimas::image< T > operator()( const mimas::image< T > &in ) const;
 protected:
  /// The tensor.
  Tensor t;
  /// Background value for undefined pixels.
  T background;
};

///@}

};

#include "warp.tcc"

#endif
