#ifndef __XML_REFERENCE_H
#define __XML_REFERENCE_H

#include "mimasconfig.h"

#ifdef HAVE_XALAN

#include <cassert>
#include <string>
#include <XalanDOM/XalanNode.hpp>
#include <util/XMLException.hpp>
#include "mimasexception.h"
#include "object.h"

#ifdef XERCES_HAS_CPP_NAMESPACE
using namespace XERCES_CPP_NAMESPACE;
#endif
#ifdef XALAN_HAS_CPP_NAMESPACE
using namespace XALAN_CPP_NAMESPACE;
#endif

namespace mimas {

class xml_document;
class xml_node_reference_list;

/** @addtogroup xml
    @{ */
/** Wrapper for XML-Nodes.
    This wrapper is used to wrap references to objects of type
    \c XalanNode and \c XalanElement.
    @see xml_document.
    @author Jan Wedekind (jan@wedesoft.de)
    @date Wed Jan 12 13:53:41 UTC 2004 */
template< class T >
class xml_reference: public object
{
 public:
  /// Constructor for internal use only.
  xml_reference( T *_node, xml_document *_document );
  /// Copy-constructor.
  xml_reference( const xml_reference< T > &r );
  /// Assignment-operator.
  xml_reference< T > &operator=( const xml_reference< T > &r );
  /** Dereferencing-operator.
      @return Reference to wrapped object. */
  T &operator*(void) const { assert( node != NULL ); return *node; }
  /** Dereferencing-operator.
      @return Reference to wrapped object. */
  T *operator->(void) const { assert( node != NULL ); return node; }
  /** Get pointer to wrapped object.
      @return Pointer to wrapped object. */
  T *get(void) const { return node; }
  /** Check pointer for being not-null.
      @return \c true, if pointer is not \c NULL. */
  operator bool(void) const { return node != NULL; }
  /** Check pointer for being null.
      @return \c true, if pointer is \c NULL. */
  bool operator! () const { return node == NULL; }
  /** Get text contained in this node.
      @return Sum of text-elements in the sub-tree. */
  std::string getNodeText(void) const throw (mimasexception,XMLException);
  /** Select node with given X-path.
      This method is <B>only implemented for T=XalanElement</B>!
      
      To convert \c xml_refence<XalanNode> to
      \c xml_refence<XalanElement> do:
      \code
      using namespace mimas;
      ...
      xml_reference< XalanElement > =
         dynamic_pointer_cast< XalanElement >( ... );
      ...
      \endcode
      @param xPath X-Path to the desired node.
      @return Wrapper containing desired node. */    
  xml_reference< XalanNode > selectNode( const char *xPath ) const
    throw (mimasexception,XMLException);
  /** Select list of nodes with given X-path.
      This method is <B>only implemented for T=XalanElement</B>!
      @param xPath X-Path for finding the desired nodes.
      @return Wrapper containing list of desired nodes. */
  xml_node_reference_list selectNodes( const char *xPath ) const
    throw (mimasexception,XMLException);
  /** Check for node with given X-path.
      This method only is implemented for XalanElement!
      The method will yield the same result as the following code:
      \code
      ( selectNodes( xPath ).getLength() > 0 )
      \endcode
      @param xPath X-Path of node(s) to check for.
      @return \c true if there is a node with such path. */
  bool existNode( const char *xPath ) const throw (XMLException);
  /** Get pointer to parent document.
      For internal use only. */
  xml_document *getDocument(void) const { return document; }
 protected:
  /// Pointer to node being wrapped.
  T *node;
  /// Pointer to parent document.
  xml_document *document;
};

/** Conversion function for XML-wrapper objects.
    This method is used for converting objects of type
    <TT>xml_reference<XalanNode></TT> to
    <TT>xml_reference<XalanElement></TT>.
    @see xml_reference */
template<class T, class U>
xml_reference<T> dynamic_pointer_cast( xml_reference<U> const & r )
{
  return xml_reference<T>( (T *)r.get(), r.getDocument() );
}

///@}

};

#include "xml_reference.tcc"

#endif

#endif
