#include <iostream>
#include <fstream>
#include <cmath>
#include <valarray>
#include "multi_array_op.h"
#include "image_fileoutput.h"

#include "registration.h"
#include "fourier.h"
#include "rotation.h"

#ifdef HAVE_LIBFFTW3

// Denis Tessier Sep 2004

using namespace boost;
using namespace std;

namespace mimas{

template< typename T >
struct _todouble: public unary_function< T, double >
{
  double operator()( const T &x ) const
  { return static_cast< double >( x ); }
};

template< typename T >
struct _tocomplex: public unary_function< T, std::complex< double > >
{
  std::complex< double > operator()( const T &x ) const
  { return std::complex< double >( x ); }
};

template< typename T >
struct _norm: public unary_function< T, double >
{
  double operator()( const T &x ) const { return norm( x ); }
};

ARRAY_OP_EX( todouble, _todouble )
ARRAY_OP_EX( tocomplex, _tocomplex )
ARRAY_OP_EX( norm, _norm )

multi_array<double, 2> highpass(multi_array<double, 2> a);
multi_array<double, 2> shift_array(multi_array<double, 2> a);


// Gets the x and y values of a shift between two images
void registration_parameters::registering_images_shift_only(image<int>& im1, image<int>& im2)
{
  multi_array< std::complex< double >,2 > f2
    ( fft( tocomplex( todouble( im1.rawData() ) ) ) );

  multi_array< std::complex< double >,2 > f1
    ( fft( tocomplex( todouble( im2.rawData() ) ) ) );

  //computation of the cross-power ratio  
  multi_array< std::complex< double >,2 > r( extents[ im1.getHeight() ][ im1.getWidth() ] );

  for(int j=0; j<im1.getWidth(); j++)
    {
      for(int i=0; i< im1.getHeight(); i++)
	{
	  if((norm(f1[i][j])==0)||(norm(f2[i][j])==0))
	    r[i][j] = 0;
	  else
	    r[i][j] = (f1[i][j] * conj(f2[i][j])) / (abs(f1[i][j]) * abs(f2[i][j]));
	}
    }
  
  multi_array< double, 2 > ir =  norm( invfft( r ) );

  // maximum of the absolute inverse fourier transform
  double max = 0.0;
  int x,y;

  for (int i=0; i<im1.getWidth(); i++)
    {
      for (int j=0; j<im1.getHeight(); j++)
	{
	  
	  if (ir[i][j] > max)
	    {
	      x = j;
	      y = i;
	      
	      max = ir[i][j];
	    }
	}
    }

  /*
  //check for other high values:

  std::cerr<<"max value is:"<< max<< std::endl; 
  for (int i=0; i<im1.getWidth(); i++)
    {
      for (int j=0; j<im1.getHeight(); j++)
	{
	  
	  if (ir[i][j] > .45*max)
	    {
	      int xx = j;
	      int yy = i;;

	      if (x > im1.getWidth()/2.0)
		xx = -(im1.getWidth() - x);
	      
	      if (y > im1.getHeight()/2.0)
		yy = -(im1.getHeight() - y);
 
	      
	      std::cerr<<"This value is another good candidate: "<< std::endl; 
	      std::cerr<<"x: "<<xx <<"\n" << "y :"<< yy <<std::endl;
	    }
	}
    }
  */

  
  // proper shift values
  int xx = x;
  int yy = y;;

  if (x > im1.getWidth()/2.0)
    xx = -(im1.getWidth() - x);

  if (y > im1.getHeight()/2.0)
    yy = -(im1.getHeight() - y);
  
  setShift(xx, yy);

}

// Gets the values of rotation angle and scaling factor between 2 images
void registration_parameters::registering_images(image<int>& im1, image<int>& im2)
{
  multi_array< std::complex< double >,2 > f1
    ( fft( tocomplex
	   (  shift_array(
			  todouble( im2.rawData() )
			  ) ) ) );
  
  multi_array< std::complex< double >,2 > f2
    ( fft( tocomplex
	   (  shift_array(
			  todouble( im1.rawData() )
		          ) ) ) );
    

  // application of an highpass filter on the absolute values of the fourier transforms
  multi_array< double ,2 > af1 = highpass( norm( f1 ) );
  multi_array< double ,2 > af2 = highpass( norm( f2 ) );
  
  // getting the values to log-polar coordonates
  double dTheta = 1.0 * M_PI / im1.getHeight();
  double b = pow(10, (log((double)im1.getWidth()) / log(10.0))/im1.getWidth());
  
  double theta;

  multi_array< double ,2 > i1( extents[ im1.getHeight() ][ im1.getWidth() ] );
  multi_array< double ,2 > i2( extents[ im1.getHeight() ][ im1.getWidth() ] );
  
  for (int i=0; i<im1.getWidth(); i++)
    {
      theta = i * dTheta;
      
      for (int j=0; j<im1.getHeight(); j++)
	{
	  double r, x, y;
	  r = pow(b, j) - 1.0;
	  
	  x = r * cos(theta) + im1.getHeight()/2.0 - 0.5;
	  y = r * sin(theta) + im1.getWidth()/2.0 - 0.5;
	  int k = (int) floor(x);
	  int l = (int) floor(y);	    
	  
	  double a00, a10, a11, a01;
	  double b00, b10, b11, b01;
	  
	  if ((k < im1.getWidth())   && (l < im1.getHeight())   && (x > 1)   && (y > 1))   {a00 = af1[k][l];     b00 = af2[k][l];}
	  else {a00 = 0.0; b00 = 0.0;}
	  if ((k+1 < im1.getWidth()) && (l < im1.getHeight())   && (x+1 > 1) && (y > 1))   {a10 = af1[k+1][l];   b10 = af2[k+1][l];}
	  else {a10 = 0.0; b10 = 0.0;}
	  if ((k+1 < im1.getWidth()) && (l+1 < im1.getHeight()) && (x+1 > 1) && (y+1 > 1)) {a11 = af1[k+1][l+1]; b11 = af2[k+1][l+1];}
	  else {a11 = 0.0; b11 = 0.0;}
	  if ((k < im1.getWidth())   && (l+1 < im1.getHeight()) && (x > 1)   && (y+1 > 1)) {a01 = af1[k][l+1];   b01 = af2[k][l+1];}
	  else {a01 = 0.0; b01 = 0.0;}
	  
	  double t = x - k;
	  double u = y - l;
	  
	  i1[i][j] = (1.0-t)*(1.0-u)*a00 + t*(1.0-u)*a10 + (1.0-t)*u*a01 + t*u*a11; 
	  i2[i][j] = (1.0-t)*(1.0-u)*b00 + t*(1.0-u)*b10 + (1.0-t)*u*b01 + t*u*b11; 
	}
    }
  
  multi_array< std::complex< double >,2 > fp1
    ( fft( tocomplex(  i1 ) ) );
  
  multi_array< std::complex< double >,2 > fp2
    ( fft( tocomplex(  i2 ) ) );
  
  multi_array< std::complex< double >,2 > r( extents[ im1.getHeight() ][ im1.getWidth() ] );

  for(int j=0; j<im1.getWidth(); j++)
    {
      for(int i=0; i< im1.getHeight(); i++)
	{
	    if((abs(fp1[i][j])==0)||(abs(fp2[i][j])==0))
	      r[i][j] = 0;
	    else
	      r[i][j] = (fp1[i][j] * conj(fp2[i][j])) / (abs(fp1[i][j]) * abs(fp2[i][j]));
	}
    }
  
  multi_array< double, 2 > ir =  norm( invfft( r ) );
  
  double max = 0.0;
  int x,y;
  
  for (int i=0; i<im1.getWidth(); i++)
    {
      for (int j=0; j<im1.getHeight(); j++)
	{
	  if (ir[i][j] > max)
	    {
	      x = j;
	      y = i;
	      
	      max = ir[i][j];
	    }
	}
    }

  
  double angle, scale;
  if (180.0*y/im1.getHeight() <= 90.0)
    {
      angle = 180.0*y/im1.getHeight();
      scale = pow(b,x);
    }
  else
    {
      angle = 180.0*y/im1.getHeight()-180;
      scale = 1.0/pow(b,x);
    }

  setScale(scale);
  setAngle(angle);
}





multi_array<double, 2> highpass(multi_array<double, 2> a)
{
  int width = a.shape()[0];
  int height = a.shape()[1];
  multi_array<double, 1> hp (extents[(width+1)*(height+1)]);
  multi_array<double, 2> b (extents[width][height]);

  for (int i=0; i<width+1; i++)
    {
      for (int j=0; j<height+1; j++)
	{
	  double x=cos(M_PI*((i-(width/2.0))*(1.0/width)))*cos(M_PI*((j-(height/2.0))*(1.0/height)));
	  hp[i*(width+1)+j]=(1.0-x)*(2.0-x);
	}
    }
  for (int i=0; i<width; i++)
    {
      for (int j=0; j<height; j++)
	{
	  b[i][j] = a[i][j] * hp[i+j*(width+1)];
	}
    }
  return b;
}

multi_array<double, 2> shift_array(multi_array<double, 2> a)
{
  int width = a.shape()[0];
  int height = a.shape()[1];
  multi_array<double, 2> b (extents[width][height]);
  
  for (int i=0; i<width; i++)
    {
      for (int j=0; j<height; j++)
	{
	  if (((i+j) % 2) == 1)
	    b[i][j] = (-1) * a[i][j];
	  else
	    b[i][j] = a[i][j];
	}
    }
  return b;
}

// creates a new image from one of the image that should match the other and append it
// the this other image to get an full image with all the information from the input images
image<int> registration (image<int>& inimage, image<int>& inimage2,reg_param *result_p , bool with_scale_and_rotation)
{
  registration_parameters p;
  image<int> rotated_image2;
  
  if(with_scale_and_rotation)
    {
      p.registering_images(inimage, inimage2);
      rotated_image2 = rotate_image(inimage2, -p.getAngle(), p.getScale(), 0);
    }
  else
    {
      rotated_image2 = inimage2;
    }

  /*
  //to check
  std::ofstream fileOut("rerotated_image.pgm", std::ios::binary);
  fileOut << rotated_image2;
  */
  
  image<int> extended_image1;
  extended_image1.init(rotated_image2.getWidth(), rotated_image2.getHeight());

  int size2 = rotated_image2.getWidth();
  int size1 = inimage.getWidth();
  int new_start = (size2-size1)/2;

  for(int i=0; i<size1; i++)
    for(int j=0; j<size1; j++)
      extended_image1.setPixel(new_start+i, new_start+j, inimage.getPixel(i,j));

  p.registering_images_shift_only(extended_image1, rotated_image2);

  image<int> image5;

  image5.init(size2+2*abs(p.getShiftx())-2*new_start, size2+2*abs(p.getShifty())-2*new_start);
  
  int x = p.getShiftx();
  int y = p.getShifty();

  /*
  cout << "angle :  " << p.getAngle() << endl;
  cout << "scale :  " << p.getScale() << endl;
  cout << "shift :  " << x << " " << y << endl;
  */
  //Transparency added by putting half of the value we assumed the image
  //pixel value is initialised to 0
  //!!!!I have eliminated (commented) conditions that seem to be optimisation code.
  //And that impair the image if we want to display them with transparency.
  //for the moment it seems to work but need to be checked 
  
  int i,j;
  for (i=0; i<rotated_image2.getWidth()+2*abs(x)-2*new_start; i++)
    {
      for (j=0; j<rotated_image2.getWidth()+2*abs(y)-2*new_start; j++)
	{
	  //if((i-x-abs(x)+new_start>=0) && (j-y-abs(y)+new_start>=0) && (i-x-abs(x)+new_start<size2) && (j-y-abs(y)+new_start<size2))
	    {
	      // if(extended_image1.getPixel(i-x-abs(x)+new_start, j-y-abs(y)+new_start)!=0)
		image5.setPixel(i,j,(int) (0.5 * extended_image1.getPixel(i-x-abs(x)+new_start, j-y-abs(y)+new_start)
				+ image5.getPixel(i,j)) );
	      /*   else if ((i-abs(x)+new_start>=0) && (j-abs(y)+new_start>=0) && (i-abs(x)+new_start<size2) && (j-abs(y)+new_start<size2))
		image5.setPixel(i,j, (int)(0.5 * rotated_image2.getPixel(i-abs(x)+new_start, j-abs(y)+new_start)
		+ image5.getPixel(i,j)));*/
	    }
	    // else if ((i-abs(x)+new_start>=0) && (j-abs(y)+new_start>=0) && (i-abs(x)+new_start<size2) && (j-abs(y)+new_start<size2))
	    image5.setPixel(i,j, (int)(.5*rotated_image2.getPixel(i-abs(x)+new_start, j-abs(y)+new_start)
			    + image5.getPixel(i,j)) );
	}
    }

  if(result_p != NULL)
    {
      *(result_p) = p.getRegParam();
    }
    
  return image5;
}

image<int> registration_with_canny(image<int>& inimage1, image<int>& inimage2,int lowThresh, int highThresh, float stdDev,
				      reg_param *result_p , bool with_scale_and_rotation)
{
  //To automatically set the size of the registration windows if this
  //one is too large
  /*std::cerr<<"the size parametre should not be bigger than"
    "any dimension (length and width) of both images."
    "It has been done reseted to its maximum possible size."<<std::endl;
  */
  //not very elegant but ...
  int regSize = inimage1.getWidth();
  if (regSize > inimage1.getHeight())
    regSize = inimage1.getHeight();
  if (regSize > inimage2.getHeight())
    regSize = inimage2.getHeight();
  if (regSize > inimage2.getWidth())
    regSize = inimage2.getWidth();
  //experimental test has shown it doesn't work
  //on an example without the -1. Further analysis
  //should be carried out,
  //!!!!!!check that if this value must be even. 
  regSize -= 1; 
  //std::cerr<<"3rd parametre has been set to:"<<regSize <<std::endl;
      
  //to enhance the image:
  //on images without too much features the registration doesn't work well
  //to improve this an edge filter (canny) is used and resultinf edges are
  //pasted on the image. This seems to work however I don't understand why
  //My guess would be that the fourier transform is somehow more different
  //and thus registration improved.
    
  image<int> edge_image1 = inimage1;
  image<int> edge_image2 = inimage2;
  canny<int> canny;
    
  canny.setCannyLowThres(lowThresh);
  canny.setCannyHighThres(highThresh);
  canny.setCannyAutoThres(false); // default is true
  canny.setCannyStdDev(stdDev);

  canny.getEdges(edge_image1);
  canny.getEdges(edge_image2);

  for (int i = 0; i < inimage1.getWidth(); i++)
    for (int j = 0; j < inimage1.getHeight(); j++)
      {
	if(edge_image1.getPixel(i,j) !=0)
	  inimage1.setPixel(i,j,0);
      }

  for (int i = 0; i < inimage2.getWidth(); i++)
    for (int j = 0; j < inimage2.getHeight(); j++)
      {
	if(edge_image2.getPixel(i,j) !=0)
	  inimage2.setPixel(i,j,0);
      }

  typedef boost::multi_array< double, 2 > array_type;
  typedef array_type::index_range range;
    
  image< int > im1, im2;
  im1.init( regSize, regSize ), im2.init( regSize, regSize );
  im1.rawData() = inimage1.rawData()[ boost::indices[ range( 0,regSize ) ]
				      [ range( 0, regSize ) ] ];
  im2.rawData() = inimage2.rawData()[ boost::indices[ range( 0, regSize ) ]
				      [ range( 0, regSize ) ] ];

  return registration(im1,im2, result_p);
}

}

#endif
